/*
 * These classes in namespace "Avi" have been written by
 * kbinani (Osaka, Japan)
 * kbinani@msn.com
 * 
 * Copyright (c) 2008, kbinani
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
#if DEBUG
using System.Windows.Forms;
#endif

namespace Boare.Lib.Avi {

    [StructLayout( LayoutKind.Sequential, Pack = 1 )]
    public struct BITMAPINFOHEADER {
        public UInt32 biSize;
        public Int32 biWidth;
        public Int32 biHeight;
        public Int16 biPlanes;
        public Int16 biBitCount;
        public UInt32 biCompression;
        public UInt32 biSizeImage;
        public Int32 biXPelsPerMeter;
        public Int32 biYPelsPerMeter;
        public UInt32 biClrUsed;
        public UInt32 biClrImportant;
    }


    public struct MainAVIHeader {
        public uint dwMicroSecPerFrame;     // specifies the period between video frames. This value indicates the overall timing for the file.
        public uint dwMaxBytesPerSec;       // specifies the approximate maximum data rate of the file.
        public uint dwReserved1;
        public uint dwFlags;                // AVIF_HASINDEX, AVIF_MUSTUSEINDEX, AVIF_ISINTERLEAVED, AVIF_WASCAPTUREFILE, AVIF_COPYRIGHTED
        public uint dwTotalFrames;          // specifies the total number of frames of data in file.
        public uint dwInitialFrames;        // used for interleaved files. If you are creating interleaved files, specify the number of frames in the file prior to the initial frame of the AVI sequence in this field.
        public uint dwStreams;              // specifies the number of streams in the file. For example, a file with audio and video has 2 streams.
        public uint dwSuggestedBufferSize;  // specifies the suggested buffer size for reading the file.
        public uint dwWidth;                // specify the width of the AVI file in pixels.
        public uint dwHeight;               // specify the height of the AVI file in pixels.
        public uint dwScale;                // used to specify the general time scale that the file will use.
        public uint dwRate;                 // used to specify the general time scale that the file will use.
        public uint dwStart;                // specify the starting time of the AVI file and the length of the file. The units are 
        public uint dwLength;               // defined by dwRate and dwScale. The dwStart field is usually set to zero.
    }


    public struct AVIStreamHeader {
        public uint fccType;//    character(len = 4) fccType
        public uint fccHandler;//character(len = 4) fccHandler
        public uint dwFlags;
        public uint dwReserved1;
        public uint dwInitialFrames;
        public uint dwScale;
        public uint dwRate;
        public uint dwStart;
        public uint dwLength;
        public uint dwSuggestedBufferSize;
        public uint dwQuality;
        public uint dwSampleSize;
    }


    public class RawAvi2Writer {

        const byte AVI_INDEX_OF_INDEXES = 0x00; //when each entry in aIndex
                                                // array points to an index chunk
        const byte AVI_INDEX_OF_CHUNKS = 0x01;  // when each entry in aIndex
                                                // array points to a chunk in the file
        const byte AVI_INDEX_IS_DATA = 0x80;    // when each entry is aIndex is
                                                // really the data
        const byte AVI_INDEX_2FIELD = 0x01;     // when fields within frames
                                                // are also indexed

        const int AVIF_HASINDEX = 16;           // Indicates the AVI file has an gidx1h chunk.
        const int AVIF_MUSTUSEINDEX = 32;       // Indicates the index should be used to determine the order of presentation of the data.
        const int AVIF_ISINTERLEAVED = 256;     // Indicates the AVI file is interleaved.
        const int AVIF_WASCAPTUREFILE = 65536;  // Indicates the AVI file is a specially allocated file used for capturing real-time video.
        const int AVIF_COPYRIGHTED = 131072;    // Indicates the AVI file contains copyrighted data.
        const int AVIF_TRUSTCKTYPE = 2048;      // Use CKType to find key frames

        public MainAVIHeader mainHeader;
        public AVIStreamHeader streamHeader;
        //long currentIndex;
        BinaryWriter fp;
        //int frameRate;
        //int noOfFrame;
        int m_current_chunk = 0;
        long m_position_in_chunk = 0L;
        const long SRESHOLD = 1000000000L; //AVIXXg̍őTCY(byte)
        long m_split_sreshold = 1000000L;  //AVIXg̍őTCY(byte)
        AVISTDINDEX m_std_index;
        AVISUPERINDEX m_super_index;
        int m_linesize;//bitmap1s̃f[^TCY(byte)
        bool m_is_first = true;
        long m_riff_position;//"RIFF*"*LĂt@C̐ΈʒuBRIFF-AVI ̂Ƃ͕K4BRIFF-AVIX͕̎ω
        long m_movi_position;
        long m_next_framedata_position;//00dbf[^Lׂꏊ
        long m_avix_position;
        int m_junk_length;
        uint m_scale;
        uint m_rate;

        public struct _avistdindex_entry {
            public UInt32 dwOffset;
            public UInt32 dwSize;
        }


        public struct AVISTDINDEX {
            public readonly string fcc;
            //private UInt32 m_cb;
            public UInt32 cb {
                get {
                    if ( aIndex == null ) {
                        return 24;
                    } else {
                        return (uint)(24 + 8 * aIndex.Count);
                    }
                }
            }
            public readonly UInt16 wLongsPerEntry;
            public readonly byte bIndexSubType;
            public readonly byte bIndexType;
            //private UInt32 m_nEntriesInUse;
            public UInt32 nEntriesInUse {
                get {
                    if ( aIndex == null ) {
                        return 0;
                    } else {
                        return (uint)aIndex.Count;
                    }
                }
            }
            public readonly string dwChunkId;
            public UInt64 qwBaseOffset;
            public readonly UInt32 dwReserved3;
            public List<_avistdindex_entry> aIndex;
            public AVISTDINDEX( ulong BaseOffset ) {
                fcc = "ix00";
                wLongsPerEntry = 2;
                bIndexSubType = 0;
                bIndexType = AVI_INDEX_OF_CHUNKS;
                dwChunkId = "00db";
                qwBaseOffset = BaseOffset;
                dwReserved3 = 0;
                //m_nEntriesInUse = 0;
                aIndex = new List<_avistdindex_entry>();
            }
            public void SetBaseOffset( UInt64 base_offset ) {
                this.qwBaseOffset = base_offset;
            }
            public void AddIndex( uint dwOffset, uint dwSize ) {
                _avistdindex_entry entry = new _avistdindex_entry();
                entry.dwOffset = dwOffset;
                entry.dwSize = dwSize;
                this.aIndex.Add( entry );
            }
            public void Write( BinaryWriter fp ) {
                fwrite( fcc, fp );
                WriteDWORD( (uint)cb, fp );
                WriteWORD( wLongsPerEntry, fp );
                WriteBYTE( bIndexSubType, fp );
                WriteBYTE( bIndexType, fp );
                WriteDWORD( (uint)nEntriesInUse, fp );
                fwrite( dwChunkId, fp );
                WriteQWORD( (long)qwBaseOffset, fp );
                WriteDWORD( (uint)dwReserved3, fp );
                foreach ( _avistdindex_entry entry in aIndex ) {
                    WriteDWORD( (uint)entry.dwOffset, fp );
                    WriteDWORD( (uint)entry.dwSize, fp );
                }
            }
        }


        public struct _avisuperindex_entry {
            public UInt64 qwOffset;
            public UInt32 dwSize;
            public UInt32 dwDuration;
        }


        public struct AVISUPERINDEX {
            public readonly string fcc;
            //public UInt32 cb;
            public UInt32 cb {
                get {
                    if ( aIndex == null ) {
                        return 24;
                    } else {
                        return (uint)(24 + 16 * aIndex.Count);
                    }
                }
            }
            public readonly UInt16 wLongsPerEntry;
            public byte bIndexSubType;
            public byte bIndexType;
            public UInt32 nEntriesInUse;
            public readonly string dwChunkId;
            public UInt32 dwReserved1;
            public UInt32 dwReserved2;
            public UInt32 dwReserved3;
            public List<_avisuperindex_entry> aIndex;
            public AVISUPERINDEX( int dumy ) {
                this.fcc = "indx";
                this.wLongsPerEntry = 4;
                this.bIndexSubType = 0;
                this.bIndexType = AVI_INDEX_OF_INDEXES;
                this.nEntriesInUse = 0;
                this.dwChunkId = "00db";
                this.dwReserved1 = 0;
                this.dwReserved2 = 0;
                this.dwReserved3 = 0;
                this.aIndex = new List<_avisuperindex_entry>();
                _avisuperindex_entry entry = new _avisuperindex_entry();
                entry.qwOffset = 0;
                entry.dwSize = 0;
                entry.dwDuration = 0;
                for ( int i = 0; i < 126; i++ ) {
                    this.aIndex.Add( entry );
                }
            }
            public void Write( BinaryWriter fp ) {
                fwrite( fcc, fp );
                WriteDWORD( (uint)cb, fp );//قƂ(int)cb
                WriteWORD( wLongsPerEntry, fp );
                WriteBYTE( bIndexSubType, fp );
                WriteBYTE( bIndexType, fp );
                WriteDWORD( (uint)nEntriesInUse, fp );
                fwrite( dwChunkId, fp );
                WriteDWORD( (uint)dwReserved1, fp );
                WriteDWORD( (uint)dwReserved2, fp );
                WriteDWORD( (uint)dwReserved3, fp );
                foreach ( _avisuperindex_entry entry in aIndex ) {
                    WriteQWORD( (long)entry.qwOffset, fp );
                    WriteDWORD( (uint)entry.dwSize, fp );
                    WriteDWORD( (uint)entry.dwDuration, fp );
                }
            }
        }


        public static uint mmioFOURCC( string fcc ) {
            char[] str = new char[4];
            for ( int i = 0; i < 4; i++ ) {
                if ( i < fcc.Length ) {
                    str[i] = fcc[i];
                } else {
                    str[i] = ' ';
                }
            }
            return mmioFOURCC( str[0], str[1], str[2], str[3] );
        }
        public static uint mmioFOURCC( char ch0, char ch1, char ch2, char ch3 ) {
            return (uint)((byte)(ch0) | ((byte)(ch1) << 8) | ((byte)(ch2) << 16) | ((byte)(ch3) << 24));
        }

        internal float frameRate {
            get {
                return (float)m_rate / (float)m_scale;
            }
        }

        /// <summary>
        /// w肵AVI_CONTAINER\̂AVIt@C̏i[ƋɁC
        /// t@CɃwb_[݂܂D
        /// </summary>
        /// <param name="file">ݑΏۂ̃t@C</param>
        /// <param name="frameRate">AVIt@C̃t[[g</param>
        public void Open( string file, uint scale, uint rate ) {
            this.fp = new BinaryWriter( new FileStream( file, FileMode.Create, FileAccess.Write ) );
            float fps = (float)rate / (float)scale;
            this.mainHeader.dwMicroSecPerFrame = (uint)(1.0e6 / fps);//  ! 1b10^6ʕb
            this.mainHeader.dwReserved1 = 0;
            this.mainHeader.dwFlags = 2064;
            this.mainHeader.dwInitialFrames = 0;
            this.mainHeader.dwStreams = 0;
            this.mainHeader.dwScale = scale;
            this.mainHeader.dwRate = rate;
            this.mainHeader.dwStart = 0;
            this.mainHeader.dwLength = 0;
            this.m_rate = rate;
            this.m_scale = scale;
            //this.noOfFrame = 0;
            fwrite( "RIFF", this.fp );
            WriteDWORD( 0, this.fp );
            fwrite( "AVI ", this.fp );
            fwrite( "LIST", this.fp );
            WriteDWORD( 0x9cc, this.fp );
            fwrite( "hdrl", this.fp );
            m_current_chunk = 0;
            m_position_in_chunk = 0L;
            m_std_index = new AVISTDINDEX( 0L );
            m_super_index = new AVISUPERINDEX( 0 );
            m_riff_position = 0x4;
        }
        [Obsolete]
        public void Open( string file, float frameRate ) {
            uint scale, rate;
            AviWriter.CalcScaleAndRate( (decimal)frameRate, out scale, out rate );
            Open( file, scale, rate );
        }


        //todo: AVIMainHeader.dwTotalFramesɁAt@CŜ̃t[idlᔽj
        //todo: AVIStreamHeader.dwLengthɁAt@CŜ̃t[idlᔽj
        /// <summary>
        /// SẴCfbNXXVAt@Cit@CƂājgpłԂɂ܂
        /// ̊֐ǂ񂾂ƂłCaviAddFrame֐gƂŃt[ǉ邱Ƃo܂D
        /// </summary>
        public void UpdateIndex() {

            _avisuperindex_entry entry = m_super_index.aIndex[m_current_chunk];
            entry.qwOffset = (ulong)fp.BaseStream.Position;
            entry.dwSize = m_std_index.cb;
            m_super_index.aIndex[m_current_chunk] = entry;
            if ( fp.BaseStream.Position != m_next_framedata_position ) {
                fp.BaseStream.Seek( m_next_framedata_position, SeekOrigin.Begin );
            }
            m_std_index.Write( fp );
            int frames = (int)m_super_index.aIndex[m_current_chunk].dwDuration;
            m_avix_position = fp.BaseStream.Position;

            if ( m_current_chunk == 0 ) {
                uint i, step, number;
                step = this.mainHeader.dwSuggestedBufferSize + 8;

                fwrite( "idx1", this.fp );
                WriteDWORD( (uint)(16 * frames), this.fp );
                for ( i = 1; i <= frames; i++ ) {
                    fwrite( "00db", this.fp );
                    WriteDWORD( AVIF_HASINDEX, this.fp );
                    WriteDWORD( (uint)(4 + (i - 1) * step), this.fp );
                    WriteDWORD( this.mainHeader.dwSuggestedBufferSize, this.fp );
                }//    end do
                m_avix_position = fp.BaseStream.Position;

                number = (uint)frames;
                //avi_writeIsolate( this.fp, number, 48 );    // AVIMainHeader.dwTotalFrames
                fp.Seek( 48, SeekOrigin.Begin );
                WriteDWORD( number, fp );
                //avi_writeIsolate( this.fp, number, 140 );   // AVIStreamHeader.dwLength
                fp.Seek( 140, SeekOrigin.Begin );
                WriteDWORD( number, fp );

                //odml dlmhdwTotalFrames
                fp.Seek( 0x8e8, SeekOrigin.Begin );
                WriteDWORD( (uint)frames, fp );

                // LIST****movi****̐lvZ
                number = 4;//"movi"
                number += (uint)(frames * (m_linesize * mainHeader.dwHeight + 8));//t[*(t[̃TCY+"00db"+00db`ÑTCY)
                number += 4 + 4 + (uint)m_std_index.cb; //ix00̃TCY
                //number += 4 + 4 + 16 * frames;//idx1̃TCY
                //avi_writeIsolate( this.fp, number, 2040 );  // LIST****movi  ****
                fp.BaseStream.Seek( m_movi_position, SeekOrigin.Begin );
                WriteDWORD( number, fp );

                //number = 4096 + (this.mainHeader.dwSuggestedBufferSize + 24) * this.noOfFrame;
                //avi_writeIsolate( this.fp, number, 4 );     // RIFF****AVI   ****
                number = (uint)m_junk_length/* 0xff4*/; //JUNK̏I܂ł̃TCYB͌Œ
                number += 4;//"movi"
                number += (uint)(frames * (m_linesize * mainHeader.dwHeight + 8));//00db...̍vTCY
                number += 4 + 4 + (uint)m_std_index.cb;
                number += (uint)(4 + 4 + 16 * frames);//idx1̃TCY
                fp.BaseStream.Seek( m_riff_position, SeekOrigin.Begin );
                WriteDWORD( number, fp );
                UpdateIndexOfIndex();
            } else {

                // LIST****movi****XV
                int number = 4;
                number += (int)(frames * (m_linesize * mainHeader.dwHeight + 8));
                number += 8 + (int)m_std_index.cb;
                fp.BaseStream.Seek( m_movi_position, SeekOrigin.Begin );
                WriteDWORD( (uint)number, fp );

                // odml dlmhdwTotalFrames
                uint frames2 = 0;
                for ( int j = 0; j <= m_current_chunk; j++ ) {
                    frames2 += (uint)m_super_index.aIndex[j].dwDuration;
                }
                fp.BaseStream.Seek( 0x8e8, SeekOrigin.Begin );
                WriteDWORD( frames2, fp );

                //avi_writeIsolate( this.fp, number, 48 );    // AVIMainHeader.dwTotalFrames
                fp.Seek( 48, SeekOrigin.Begin );
                WriteDWORD( frames2, fp );

                //avi_writeIsolate( this.fp, number, 140 );   // AVIStreamHeader.dwLength
                fp.Seek( 140, SeekOrigin.Begin );
                WriteDWORD( frames2, fp );

                //RIFF****AVIX****XV
                long num2 = m_junk_length + number;
                fp.BaseStream.Seek( m_riff_position, SeekOrigin.Begin );
                WriteDWORD( (uint)num2, fp );
                UpdateIndexOfIndex();
            }

        }


        /// <summary>
        /// avit@C܂
        /// </summary>
        public void Close() {
            UpdateIndex();
            this.fp.Close();
        }


        /// <summary>
        /// ŏ[AVI :AVI[LIST:hdrl[LIST:strl]]]ɏ܂Ăsuper index`N[indx]XV܂
        /// </summary>
        private void UpdateIndexOfIndex() {
            fp.Seek( 0xd4, SeekOrigin.Begin );
            m_super_index.Write( fp );
        }


        /// <summary>
        /// avit@CɃt[1ǉ܂D
        /// </summary>
        /// <param name="bmp"></param>
        public void AddFrame( Bitmap bmp ) {
            int i, width, height, lineSize;

            // BitmapDatarbg}bvf[^ƁABITMPAINFOHEADERo
            BitmapData bmpDat = bmp.LockBits(
                    new Rectangle( 0, 0, (int)bmp.Width, (int)bmp.Height ), ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb );
            int address = bmpDat.Scan0.ToInt32();
            byte[] bitmapData = new byte[bmpDat.Stride * bmpDat.Height];
            Marshal.Copy( new IntPtr( address ), bitmapData, 0, bitmapData.Length );

            if ( m_is_first ) {//then
                m_is_first = false;
                this.mainHeader.dwWidth = (uint)bmpDat.Width;// bmp%infoHeader%Width
                this.mainHeader.dwHeight = (uint)bmpDat.Height;// bmp%infoHeader%Height
                this.mainHeader.dwMaxBytesPerSec = (uint)(bmpDat.Stride * bmpDat.Height * this.frameRate);// bmp%infoHeader%SizeImage * avi%frameRate
                this.mainHeader.dwStreams = 1;
                this.mainHeader.dwSuggestedBufferSize = (uint)(bmpDat.Stride * bmpDat.Height);// bmp.infoHeader%SizeImage
                m_linesize = bmpDat.Stride;

                this.streamHeader.fccType = mmioFOURCC( "vids" );
                this.streamHeader.fccHandler = 0;
                this.streamHeader.dwFlags = 0;
                this.streamHeader.dwReserved1 = 0;
                this.streamHeader.dwInitialFrames = 0;
                this.streamHeader.dwScale = m_scale;
                this.streamHeader.dwRate = m_rate;
                this.streamHeader.dwStart = 0;
                this.streamHeader.dwSuggestedBufferSize = this.mainHeader.dwSuggestedBufferSize;
                this.streamHeader.dwQuality = 0;
                this.streamHeader.dwSampleSize = 0;

                aviWriteMainHeader();

                fwrite( "LIST", this.fp );// i = fwrite( 'LIST', 1, 4, avi%fp )
                WriteDWORD( 0x874, this.fp );// call bmpQWordWrite( 130, avi%fp )
                fwrite( "strl", this.fp );// i = fwrite( 'strl', 1, 4, avi%fp )

                aviWriteStreamHeader();// avi )

                fwrite( "strf", this.fp );// i = fwrite( 'strf', 1, 4, avi%fp )
                WriteDWORD( 0x28, this.fp ); //call bmpQWordWrite( 40, avi%fp )    !// infoHeader̃TCY
                BITMAPINFOHEADER bih = new BITMAPINFOHEADER();
                bih.biSize = (uint)(Marshal.SizeOf( bih ));
                bih.biWidth = bmpDat.Width;
                bih.biHeight = bmpDat.Height;
                bih.biPlanes = 1;
                bih.biBitCount = 24;
                bih.biCompression = 0;//BI_RGB
                bih.biSizeImage = (uint)(bmpDat.Stride * bmpDat.Height);
                bih.biXPelsPerMeter = 0;
                bih.biYPelsPerMeter = 0;
                bih.biClrUsed = 0;
                bih.biClrImportant = 0;
                bmpWriteInfoHeader( bih, this.fp );

                /*fwrite( "strn", this.fp );
                WriteDWORD( 6, this.fp );
                fwrite( "VIDEO", this.fp );
                WriteBYTE( 0, this.fp );*/
                fwrite( "indx", this.fp );          //fcc
                WriteDWORD( 0x7f8, this.fp );       // cb
                WriteWORD( (byte)0x4, this.fp );    // wLongsPerEntry
                WriteBYTE( 0x0, this.fp );          // bIndexSubType
                WriteBYTE( AVI_INDEX_OF_INDEXES, this.fp );// bIndexType
                WriteDWORD( 0x0, this.fp );         // nEntriesInUse
                fwrite( "00db", this.fp );          // dwChunkId
                WriteDWORD( 0x0, this.fp );
                WriteDWORD( 0x0, this.fp );
                WriteDWORD( 0x0, this.fp );
                for ( int ii = 1; ii <= 126; ii++ ) {
                    WriteQWORD( 0x0, this.fp );
                    WriteDWORD( 0x0, this.fp );
                    WriteDWORD( 0x0, this.fp );
                }

                fwrite( "LIST", this.fp );
                WriteDWORD( 0x104, fp );
                fwrite( "odml", this.fp );
                fwrite( "dmlh", fp );
                WriteDWORD( 0xf8, fp );
                WriteDWORD( 0x0, fp );//ōXVׂ
                for ( int ii = 1; ii <= 61; ii++ ) {
                    WriteDWORD( 0x0, fp );
                }

                fwrite( "JUNK", this.fp );// i = fwrite( 'JUNK', 1, 4, avi%fp )
                WriteDWORD( 0x60c, fp );
                WriteDWORD( 0, fp );//"This"o^ꂽ΂̂
                fwrite( "This file was generated by RawAvi@LipSync", this.fp );
                //WriteDWORD( 1503, this.fp );// call bmpQWordWrite( 1802, avi%fp )
                for ( i = 1; i <= 1503; i++ ) {//do i = 1, 1802
                    this.fp.Write( (byte)0 );// call fputc( 0, avi%fp )
                }//end do
                m_junk_length = 0xff4;

                fwrite( "LIST", this.fp );//      i = fwrite( 'LIST', 1, 4, avi%fp )
                m_movi_position = fp.BaseStream.Position;
                WriteDWORD( 0, this.fp );// call bmpQWordWrite( 0, avi%fp )     !// ******************RR͈̐ԍŌɏKv2040`2043dwTotalFramesi48`51j
                fwrite( "movi", this.fp );// i = fwrite( 'movi', 1, 4, avi%fp )
                m_next_framedata_position = fp.BaseStream.Position;

                m_std_index.SetBaseOffset( (ulong)m_next_framedata_position );
                m_super_index.nEntriesInUse++;
            }//end if

            if ( m_next_framedata_position != fp.BaseStream.Position ) {
                fp.BaseStream.Seek( m_next_framedata_position, SeekOrigin.Begin );
            }

            long chunk_size = m_next_framedata_position - m_riff_position;
#if DEBUG
//            MessageBox.Show( "chunk_size=" + chunk_size );
#endif
            if ( (m_current_chunk == 0 && chunk_size > m_split_sreshold) ||
                 (m_current_chunk > 0 && chunk_size > SRESHOLD )) {
                // AVIXXgւ݂̏Ɉڍs
                UpdateIndex();
                fp.BaseStream.Seek( m_avix_position, SeekOrigin.Begin );
                fwrite( "RIFF", fp );
                m_riff_position = fp.BaseStream.Position;
#if DEBUG
//                fp.Flush();
//                MessageBox.Show( "m_riff_position=" + m_riff_position );
#endif
                WriteDWORD( 0, fp );
                fwrite( "AVIX", fp );
                long current = fp.BaseStream.Position;
                if ( (current + 12) % 0x800 != 0 ) {
                    long additional = (current + 20) % 0x800;
                    additional = 0x800 - ((current + 20) % 0x800);
                    m_junk_length = (int)additional + 20;
                    fwrite( "JUNK", fp );
                    WriteDWORD( (uint)additional, fp );
                    for ( long ii = 0; ii < additional; ii++ ) {
                        WriteBYTE( (byte)0, fp );
                    }
                } else {
                    m_junk_length = 0;
                }
                m_junk_length = 0;

                fwrite( "LIST", fp );
                m_movi_position = fp.BaseStream.Position;
                WriteDWORD( 0, fp );//ōXVׂ
                fwrite( "movi", fp );
                m_next_framedata_position = fp.BaseStream.Position;
                m_std_index.aIndex.Clear();
                m_std_index.SetBaseOffset( (ulong)m_next_framedata_position );
                m_current_chunk++;
                m_super_index.nEntriesInUse++;
            }

            // t[ޏ
            width = (int)this.mainHeader.dwWidth;
            height = (int)this.mainHeader.dwHeight;
            if ( width != bmpDat.Width ) {//then
                //aviAddFrame = -1
                return;
            }//end if
            if ( height != bmpDat.Height ) {//then
                //aviAddframe = -1
                return;
            }//end if
            lineSize = bmpDat.Stride;// int( (width * 24 + 31) / 32 ) * 4

            m_std_index.AddIndex( (uint)((ulong)fp.BaseStream.Position - m_std_index.qwBaseOffset) + 8, (uint)(lineSize * height) );
            fwrite( "00db", this.fp );//    i = fwrite( '00db', 1, 4, avi%fp )
            WriteDWORD( mainHeader.dwSuggestedBufferSize, fp );// call bmpQWordWrite( avi%mainHeader%dwSuggestedBufferSize, avi%fp )
            fp.Write( bitmapData );
            m_next_framedata_position = fp.BaseStream.Position;
            _avisuperindex_entry entry = m_super_index.aIndex[m_current_chunk];
            entry.dwDuration++;
            m_super_index.aIndex[m_current_chunk] = entry;//    avi%noOfFrame = avi%noOfFrame + 1
            this.fp.Flush();// aviAddFrame = fflush( avi%fp )
            bmp.UnlockBits( bmpDat );
        }//  end function


        /// <summary>
        /// w肳ꂽBITMAP^ϐ̏wb_[t@Cɏ݂܂D
        /// </summary>
        /// <param name="bmp"></param>
        /// <param name="fp"></param>
        private static void bmpWriteInfoHeader( BITMAPINFOHEADER infoHeader, BinaryWriter fp ) {
            //type(INFO_HEADER), intent(in) :: infoHeader
            //type(FILE), intent(inout) :: fp
            WriteDWORD( (uint)infoHeader.biSize, fp );
            WriteDWORD( (uint)infoHeader.biWidth, fp );
            WriteDWORD( (uint)infoHeader.biHeight, fp );
            WriteWORD( (uint)(infoHeader.biPlanes), fp );
            WriteWORD( (uint)(infoHeader.biBitCount), fp );
            WriteDWORD( (uint)infoHeader.biCompression, fp );
            WriteDWORD( (uint)infoHeader.biSizeImage, fp );
            WriteDWORD( (uint)infoHeader.biXPelsPerMeter, fp );
            WriteDWORD( (uint)infoHeader.biYPelsPerMeter, fp );
            WriteDWORD( (uint)infoHeader.biClrUsed, fp );
            WriteDWORD( (uint)infoHeader.biClrImportant, fp );
        }//end subroutine


        /// <summary>
        /// t@CMainAviHeader\̂̒l݂܂
        /// </summary>
        private void aviWriteMainHeader() {
            //type(AVI_CONTAINER), intent(inout) :: avi
            fwrite( "avih", this.fp );//    i = fwrite( 'avih', 1, 4, avi%fp )
            WriteDWORD( 56, this.fp );    // MainAVIHeader̃TCY
            WriteDWORD( this.mainHeader.dwMicroSecPerFrame, this.fp );
            WriteDWORD( 0/*this.mainHeader.dwMaxBytesPerSec*/, this.fp );
            WriteDWORD( this.mainHeader.dwReserved1, this.fp );
            WriteDWORD( this.mainHeader.dwFlags, this.fp );
            WriteDWORD( this.mainHeader.dwTotalFrames, this.fp );
            WriteDWORD( this.mainHeader.dwInitialFrames, this.fp );
            WriteDWORD( this.mainHeader.dwStreams, this.fp );
            WriteDWORD( 0/*this.mainHeader.dwSuggestedBufferSize*/, this.fp );
            WriteDWORD( this.mainHeader.dwWidth, this.fp );
            WriteDWORD( this.mainHeader.dwHeight, this.fp );
            WriteDWORD( this.mainHeader.dwScale, this.fp );
            WriteDWORD( this.mainHeader.dwRate, this.fp );
            WriteDWORD( this.mainHeader.dwStart, this.fp );
            WriteDWORD( this.mainHeader.dwLength, this.fp );
        }//end subroutine


        /// <summary>
        /// t@CAVIStreamHeader\̂̒l݂܂
        /// </summary>
        private void aviWriteStreamHeader() {
            //type(AVI_CONTAINER), intent(inout) :: avi
            fwrite( "strh", fp );
            WriteDWORD( 56, this.fp );// call bmpQWordWrite( 56, avi%fp )    !// AVIStreamHeader̃TCY
            //fwrite( this.streamHeader.fccType, this.fp );// i = fwrite( avi%streamHeader%fccType, 1, 4, avi%fp )
            WriteDWORD( (uint)streamHeader.fccType, fp );
            //fwrite( this.streamHeader.fccHandler, this.fp );//            i = fwrite( this.streamHeader.fccHandler, 1, 4, this.fp );
            WriteDWORD( (uint)streamHeader.fccHandler, fp );
            //WriteDWORD( 0, this.fp );
            WriteDWORD( this.streamHeader.dwFlags, this.fp );
            //WriteDWORD( this.streamHeader.dwReserved1, this.fp );
            WriteWORD( 0, this.fp );//wPriority
            WriteWORD( 0, this.fp );//wLanghage
            WriteDWORD( this.streamHeader.dwInitialFrames, this.fp );
            WriteDWORD( this.streamHeader.dwScale, this.fp );
            WriteDWORD( this.streamHeader.dwRate, this.fp );
            WriteDWORD( this.streamHeader.dwStart, this.fp );
            WriteDWORD( this.streamHeader.dwLength, this.fp );
            WriteDWORD( this.streamHeader.dwSuggestedBufferSize, this.fp );
            WriteDWORD( this.streamHeader.dwQuality, this.fp );
            WriteDWORD( this.streamHeader.dwSampleSize, this.fp );
            WriteWORD( 0, this.fp );//left
            WriteWORD( 0, this.fp );//top
            WriteWORD( mainHeader.dwWidth, this.fp );//right
            WriteWORD( mainHeader.dwHeight, this.fp );//bottom
        }//end subroutine


        /// <summary>
        /// w肵ʒunumber̕\4oCg݂܂
        /// </summary>
        /// <param name="fp"></param>
        /// <param name="number">ސ</param>
        /// <param name="start">݂Jnʒu</param>
        private static void DESTROYED_avi_writeIsolate( BinaryWriter fp, int number, int start ) {
            long current;//integer(DPI) current
            int k1, k2, k3, k4, tmp;//integer status, k1, k2, k3, k4, tmp
            current = fp.BaseStream.Position;// current = ftell( fp )
            //fp.Seek( start - 1, SeekOrigin.Begin );// status = fseek( fp, start - 1, SEEK_SET )
            fp.Seek( start, SeekOrigin.Begin );// status = fseek( fp, start - 1, SEEK_SET )
            k1 = number % 256;//    k1  = mod( number, 256 )
            tmp = (number - k1) / 256;
            k2 = tmp % 256;// mod( tmp, 256 )
            tmp = (tmp - k2) / 256;
            k3 = tmp % 256;// mod( tmp, 256 )
            tmp = (tmp - k3) / 256;
            k4 = tmp % 256;// mod( tmp, 256 )
            fp.Write( (byte)k1 );
            fp.Write( (byte)k2 );
            fp.Write( (byte)k3 );
            fp.Write( (byte)k4 );
            fp.Flush();//    status = fflush( fp )
            fp.Seek( (int)current, SeekOrigin.Begin );// status = fseek( fp, current, SEEK_SET )
        }//end subroutine


        private static void fwrite( string str, BinaryWriter fp ) {
            int length = str.Length;
            if ( length <= 0 ) {
                return;
            }
            foreach ( char ch in str ) {
                fp.Write( (byte)ch );
            }
        }


        /// <summary>
        /// BYTEl1bytet@Cɏ݂܂D
        /// </summary>
        /// <param name="number"></param>
        /// <param name="fp"></param>
        private static void WriteBYTE( byte number, BinaryWriter fp ) {
            fp.Write( number );
        }


        /// <summary>
        /// integer(2)DWORDl2bytet@Cɏ݂܂D
        /// </summary>
        /// <param name="number"></param>
        /// <param name="fp"></param>
        private static void WriteWORD( uint number, BinaryWriter fp ) {
            uint tmp, k1, k2;
            k1 = number % 256;//mod(number, 256)
            tmp = (number - k1) / 256;
            k2 = tmp % 256;// mod(tmp, 256)
            fp.Write( (byte)k1 );
            fp.Write( (byte)k2 );
        }//end subroutine

        
        /// <summary>
        /// integer(4)DWORDl4bytet@Cɏ݂܂
        /// </summary>
        /// <param name="number"></param>
        /// <param name="fp"></param>
        private static void WriteDWORD( uint number, BinaryWriter fp ) {
            uint tmp, k1, k2, k3, k4;
            k1 = number % 256;// mod(number, 256)
            tmp = (number - k1) / 256;
            k2 = tmp % 256;//mod(tmp, 256)
            tmp = (tmp - k2) / 256;
            k3 = tmp % 256;// mod(tmp, 256)
            tmp = (tmp - k3) / 256;
            k4 = tmp % 256;// mod(tmp, 256)
            fp.Write( (byte)k1 );
            fp.Write( (byte)k2 );
            fp.Write( (byte)k3 );
            fp.Write( (byte)k4 );
        }


        /// <summary>
        /// integer(8)QWORDl8bytet@Cɏ݂܂
        /// </summary>
        /// <param name="number"></param>
        /// <param name="fp"></param>
        private static void WriteQWORD( long number, BinaryWriter fp ) {
            long tmp;
            byte k1, k2, k3, k4, k5, k6, k7, k8;
            k1 = (byte)(number % 256L);// mod(number, 256)
            tmp = (number - k1) / 256;
            k2 = (byte)(tmp % 256L);//mod(tmp, 256)
            tmp = (tmp - k2) / 256;
            k3 = (byte)(tmp % 256L);//mod(tmp, 256)
            tmp = (tmp - k3) / 256;
            k4 = (byte)(tmp % 256L);//mod(tmp, 256)
            tmp = (tmp - k4) / 256;
            k5 = (byte)(tmp % 256L);//mod(tmp, 256)
            tmp = (tmp - k5) / 256;
            k6 = (byte)(tmp % 256L);//mod(tmp, 256)
            tmp = (tmp - k6) / 256;
            k7 = (byte)(tmp % 256L);//mod(tmp, 256)
            tmp = (tmp - k7) / 256;
            k8 = (byte)(tmp % 256L);//mod(tmp, 256)
            fp.Write( k1 );
            fp.Write( k2 );
            fp.Write( k3 );
            fp.Write( k4 );
            fp.Write( k5 );
            fp.Write( k6 );
            fp.Write( k7 );
            fp.Write( k8 );
        }

    }
}

